/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Microsoft Exchange Calendar Provider code.
 *
 * The Initial Developer of the Original Code is
 *   Andrea Bittau <a.bittau@cs.ucl.ac.uk>, University College London
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const CLASS_ID    = Components.ID("{201a6b60-1dfc-11de-8c30-0800200c9a66}");
const CLASS_NAME  = "Microsoft Exchange Calendar Data Provider";
const CONTRACT_ID = "@mozilla.org/calendar/calendar;1?type=exchange";

var _tbver = 0;

var calExchangeFactory = {
	createInstance: function(aOuter, aIID)
	{
		if (aOuter != null)
			throw Components.results.NS_ERROR_NO_AGGREGATION;

		return (new calExchange()).QueryInterface(aIID);
	}
};

var calExchangeModule = {
	mUtilsLoaded: false,    

	registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
	{
		aCompMgr = aCompMgr.QueryInterface(
			Components.interfaces.nsIComponentRegistrar);

		aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME,
						 CONTRACT_ID, aFileSpec,
						 aLocation, aType);
	},

	unregisterSelf: function(aCompMgr, aLocation, aType)
	{
		aCompMgr = aCompMgr.QueryInterface(
			Components.interfaces.nsIComponentRegistrar);

		aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
	},

	getClassObject: function(aCompMgr, aCID, aIID)
	{
		if (!aIID.equals(Components.interfaces.nsIFactory))
			throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

		this.loadUtils();

		if (aCID.equals(CLASS_ID))
			return calExchangeFactory;

		throw Components.results.NS_ERROR_NO_INTERFACE;
	},

	canUnload: function(aCompMgr)
	{
		return true;
	},

	loadUtils: function()
	{
		if (this.mUtilsLoaded)
			return;

		this.mScripts = ["compat.js", "calExchange.js"];

		this.getVersion();

		if (_tbver == 2)
			this.doLoadUtils2();
		else
			this.doLoadUtils3();

		this.mUtilsLoaded = true;
	},

	getVersion: function()
	{
		_tbver = 3;

		var appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
                            .getService(Components.interfaces.nsIXULAppInfo);

		var ver = appInfo.version[0];

		_tbver = parseInt(ver, 10);
//		dump("Version: " + _tbver + "\n");

		if (_tbver == 0)
			_tbver = 2;
	},

	doLoadUtils3: function()
	{
		Components.utils.import("resource://calendar/modules/calUtils.jsm");
		Components.utils.import("resource://calendar/modules/calProviderUtils.jsm");
		Components.utils.import("resource://calendar/modules/calAuthUtils.jsm");
		cal.loadScripts(["calUtils.js"], this.__parent__);

		let thisDir = __LOCATION__.parent.parent.clone();
		thisDir.append("js");

		cal.loadScripts(this.mScripts,
				this.__parent__,
				thisDir);
	},

	doLoadUtils2: function()
	{
		this.mScripts = ["compat2.js"].concat(this.mScripts);

		const kLIGHTNING_UID = "{e2fda1a4-762b-4020-b5ad-a41df1933103}";
		const baseScripts = ["calUtils.js", "calAuthUtils.js",
				     "calProviderBase.js",
				     "calProviderUtils.js"];

		var extman = Components.classes["@mozilla.org/extensions/manager;1"]
			      .getService(Components.interfaces.nsIExtensionManager);
		try {
			var appdir = extman.getInstallLocation(kLIGHTNING_UID)
				      .getItemLocation(kLIGHTNING_UID);
			appdir.append("js");
		} catch (e) {
			Components.utils.reportError("Error getting Lightning Extension Directory");
			throw(e);
		}

		for each (var script in baseScripts) {
			var scriptFile = appdir.clone();
			scriptFile.append(script);
	    
			ComponentsUtilsImport(scriptFile);
		}

		appdir = __LOCATION__.parent.parent;
		appdir.append("js");

		for each (var script in this.mScripts) {
			var scriptFile = appdir.clone();
			scriptFile.append(script);

			ComponentsUtilsImport(scriptFile);
		}
	}
};

// XXX v2 only
function ComponentsUtilsImport(aFile) {
    var iosvc = Components.classes["@mozilla.org/network/io-service;1"]
                          .getService(Components.interfaces.nsIIOService);

    var loader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                           .getService(Components.interfaces.mozIJSSubScriptLoader);

    try {
        var scriptUri = iosvc.newFileURI(aFile);
        loader.loadSubScript(scriptUri.spec, null);
    } catch (e) {
        Components.utils.reportError("Error while loading " + aFile.path);
        throw e;
    }
}

function NSGetModule(aComponentManager, aFileSpec)
{
	return calExchangeModule;
}
