/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * 
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Microsoft Exchange Calendar Provider code.
 *
 * The Initial Developer of the Original Code is
 *   Andrea Bittau <a.bittau@cs.ucl.ac.uk>, University College London
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *      
 * ***** END LICENSE BLOCK ***** */

// URI format for setting up calendar:
// http://myemail@address.com@auto for auto discovery
//
// https://myemail@address.com@myexchangeserver.com/EWS/Exchange.asmx for
// hardcoded URL.
//
// myemail@address.com is the username to log in the Exchange server.

const _mst = "http://schemas.microsoft.com/exchange/services/2006/types";
const _msm = "http://schemas.microsoft.com/exchange/services/2006/messages";

const _attendeeMap = {
	"Unknown"	: "NEEDS-ACTION",
	"Tentative"	: "TENTATIVE",
	"Accept"	: "ACCEPTED",
	"Decline"	: "DECLINED",
	"Organizer"	: "ACCEPTED"
};

const _attendeeMapRole = {
	"RequiredAttendees"	: "REQ-PARTICIPANT",
	"OptionalAttendees"	: "OPT-PARTICIPANT"
};

var _fields = new Array();

function exField(aExName, aExType, aCalName, aProperty)
{
	this.mExName   = aExName;
	this.mExType   = aExType;
	this.mCalName  = aCalName;
	this.mProperty = aProperty;
}

exField.prototype = {
	conv_e2c: function(aVal)
	{
		return aVal;
	},

	conv_c2e: function(aVal)
	{
		return aVal;
	},

	fillCal: function(aItem, aXml)
	{
		var val = getXml(aXml, this.mExName, _mst);

		val = this.conv_e2c(val);

		if (this.mProperty)
			aItem.setProperty(this.mCalName, val);
		else
			aItem[this.mCalName] = val;
	},

	doCreate: function(aItem)
	{
		if (this.mRO)
			return "";

		var val = this.val(aItem);

		if (val == null)
			return "";

		return makeTag("t:" + this.mExName, val, this.mExAttrib);
	},

	doUpdate: function(aNewItem, aOldItem)
	{
		if (this.mRO)
			return "";

		if (!this.changed(aNewItem, aOldItem))
			return "";

		var data = this.doCreate(aNewItem);

		if (data == "")
			data = makeDeleteItemField(this.mExType);
		else
			data = makeSetItemField(this.mExType, data);

		return data;
	},

	changed: function(aNewItem, aOldItem)
	{
		var a = this.doCreate(aOldItem);
		var b = this.doCreate(aNewItem);

		return a != b;
	},

	val: function(aItem)
	{
		var val = null;

		if (this.mProperty)
			val = aItem.getProperty(this.mCalName);
		else
			val = aItem[this.mCalName];

		val = this.conv_c2e(val);

		return val;
	},

	mExName: null,
	mExType: null,
	mExAttrib: null,
	mCalName: null,
	mRO: false,
	mProperty: null
};

function fillCategory(aItem, aXml)
{
	var cat = new Array();

	var i = doXPath(aXml, ".//t:Categories/t:String");

	var item;

	while ((item = i.iterateNext())) {
		cat.push(item.textContent);
	}

	if (cat.length == 0)
		return;

	aItem.setCategories(cat.length, cat);
}

function createCategory(aItem)
{
	var cats = aItem.getCategories({});

	var xml = "<t:Categories>";

	for (i in cats) {
		xml += makeTag("t:String", cats[i]);
	}

	xml += "</t:Categories>";

	return xml;
}

function fillAttendee(aItem, aXml)
{
	var attendees = aXml.getElementsByTagNameNS(_mst, this.mExName);

	if (attendees.length == 0)
		return;

	var dudes = attendees[0].getElementsByTagNameNS(_mst, "Attendee");

	var role = _attendeeMapRole[this.mExName];

	for (i = 0; i < dudes.length; i++)
		doAttendee(aItem, dudes[i], role);
}

function doAttendee(aItem, aXml, aRole, aOrganizer)
{
	if (aOrganizer == null)
		aOrganizer = false;

	let attendee = createAttendee();

	attendee.id = "mailto:" + getXml(aXml, "EmailAddress", _mst);
	attendee.commonName  	      = getXml(aXml, "Name", _mst);
	attendee.rsvp	      	      = "FALSE";
	attendee.userType    	      = "INDIVIDUAL";
	attendee.isOrganizer 	      = aOrganizer;
	attendee.role		      = aRole;
	attendee.participationStatus  = 
		_attendeeMap[getXml(aXml, "ResponseType", _mst)];

	// XXX free/busy doesn't work properly becase calid seems to tokenize on
	// commas.
	attendee.commonName = attendee.commonName.replace(",", ";");

	if (attendee.isOrganizer)
		aItem.organizer = attendee;
	else
		aItem.addAttendee(attendee);
}

function revmap(aArray, aItem)
{
	for (i in aArray) {
		if (aArray[i] == aItem)
			return i;
	}

	return null;
}

function exCreateAttendee(aItem)
{
	var attendees = aItem.getAttendees({});

	var inner = "";

	for (i in attendees) {
		var attendee = attendees[i];

		var tag = revmap(_attendeeMapRole, attendee.role);
		if (tag != this.mExName)
			continue;

		inner += doCreateAttendee(attendee);
	}

	if (inner == "")
		return "";

	var xml  = "<t:" + this.mExName + ">";
	xml 	+= inner;
	xml 	+= "</t:" + this.mExName + ">";

	return xml;
}

function doCreateAttendee(aAttendee)
{
	var xml = "";

	xml += "<t:Attendee>";

	xml += doCreateAttendeeMbox(aAttendee);

	var resp = revmap(_attendeeMap, aAttendee.participationStatus);
	xml += makeTag("t:ResponseType", resp);

	xml += "</t:Attendee>";

	return xml;
}

function doCreateAttendeeMbox(aAttendee)
{
	var xml = "";

	xml += "  <t:Mailbox>";

	var cn = aAttendee.commonName;

	// XXX
	cn = cn.replace(";", ",");

	xml += makeTag("t:Name", cn);
	xml += makeTag("t:EmailAddress", aAttendee.id.replace(/^mailto:/, ""));

	xml += "  </t:Mailbox>";

	return xml;
}

function fillOrganizer(aItem, aXml)
{
	var org = aXml.getElementsByTagNameNS(_mst, this.mExName);

	if (org.length == 0)
		return;

	doAttendee(aItem, org[0], "REQ-PARTICIPANT", true);
}

function fillMyResponse(aItem, aXml)
{
	var val = getXml(aXml, this.mExName, _mst);

	if (val == "")
		val = "Unknown";

	var attendee = this.mCal.getInvitedAttendee(aItem);

	if (!attendee)
		return;

	attendee.participationStatus = _attendeeMap[val];
}

function createMyResponse(aItem)
{
	var attendee = this.mCal.getInvitedAttendee(aItem);

	if (!attendee || (attendee == aItem.organizer))
		return "";

	var stat = revmap(_attendeeMap, attendee.participationStatus);

	return makeTag("t:" + this.mExName, stat);
}

function init_fields()
{
	add_field("Subject", "item", "title", false);

	var f = new exField("Body", "item:Body", "DESCRIPTION", true);
	f.mExAttrib = 'BodyType="Text"';
	_fields.push(f);

	f = add_field("Categories", "item", "XXX", true);
	f.fillCal  = fillCategory;
	f.doCreate = createCategory;

	init_date("Start", "calendar:Start", "startDate", false);
	init_date("End", "calendar:End", "endDate", false);

	f = init_date("DateTimeReceived", "item:DateTimeReceived",
		      "DTSTAMP", true);
	f.mRO = true;

	add_field("Location", "calendar", "LOCATION", true);

	f = add_field("Organizer", "calendar", "XXX", true);
	f.fillCal  = fillOrganizer;
	f.mRO = true;

	for (i in _attendeeMapRole) {
		f = add_field(i, "calendar", "XXX", true);
		f.fillCal		= fillAttendee;
		f.doCreate		= exCreateAttendee;
	}

	f = add_field("MyResponseType", "calendar", "XXX", true);
	f.fillCal  = fillMyResponse;
	f.mRO = true;
}

function init_date(aEx, aExType, aCal, aProp)
{
	var f = new exField(aEx, aExType, aCal, aProp);

	f.conv_e2c = _f["fromRFC3339"];
	f.conv_c2e = _f["toRFC3339"];

	_fields.push(f);

	return f;
}

function add_field(aEx, aType, aCal, aProp)
{
	var exName = aEx;
	var exType = aType + ":" + aEx;

	var f = new exField(exName, exType, aCal, aProp);

	_fields.push(f);

	return f;
}

function calExchange()
{
	this.initProviderBase();
	this.mReqQueue = new Array();
	this.mRunning = false;
}

calExchange.prototype = {
	__proto__: _proto,

	QueryInterface: function(aIID)
	{
		return doQueryInterface(this, calExchange.prototype, aIID,
					[Components.interfaces.calICalendarProvider,
					 Components.interfaces.calISchedulingSupport,
					 Components.interfaces.calIFreeBusyProvider]);
	},

	get type()
	{
		return "exchange";
	},

	set uri(aUri)
	{
		this.mUri = aUri;

		this.mUser = decodeURIComponent(aUri.username);

		sdbg("User:" + this.mUser + "\n");

		if (aUri.host != "auto") {
			this.mServer = decodeURIComponent(aUri.spec);
			sdbg("Server: " + this.mServer + "\n");
		}

		return aUri;
	},

	get uri()
	{
		return this.mUri;
	},

	getProperty: function(aName)
	{
		switch (aName) {
		case "capabilities.tasks.supported":
			return false;
		}

		return this.__proto__.__proto__.getProperty.apply(this,
								  arguments);
	},

	get canRefresh()
	{
		return true;
	},

	canNotify: function(aMethod, aItem)
	{
		return true;
	},

	init: function()
	{
		if (this.mInit != 0)
			return;

		this.mInit = 1;
		sdbg("Init\n");

		getFreeBusyService().addProvider(this);

		var tmp = this;

		let timerCallback = {
			notify: function setTimeout_notify() {
				tmp.doInit();
			}
		};

		let timer = Components.classes["@mozilla.org/timer;1"]
				.createInstance(Components.interfaces.nsITimer);

		timer.initWithCallback(timerCallback, 0, timer.TYPE_ONE_SHOT);
	},

	doInit: function()
	{
		sdbg("do init\n");
		this.getAccount();
		this.newXmlReq();

		if (this.mServer == null)
			this.autodiscover();

		this.mInit = 2;
		this.kick();
		sdbg("do init done\n");
	},

	getItem: function (aId, aListener)
	{
		sdbg("XXX TODO - this is never called anyway\n");
	},

	getItems: function(aItemFilter, aCount, aRangeStart, aRangeEnd,
			   aListener)
	{
		this.init();

		var wantInvitations = 
			((aItemFilter & Components.interfaces.calICalendar
		  		.ITEM_FILTER_REQUEST_NEEDS_ACTION) != 0);

		if (!(aItemFilter & Components.interfaces
				.calICalendar.ITEM_FILTER_TYPE_EVENT)
                    && !wantInvitations) {
			this.completeGetRequest(aListener);
			return;
		}

		var start = getDate(aRangeStart);
		var end   = getDate(aRangeEnd);

		// XXX optimize invitation - we don't need 2 calls because we
		// don't need body.  -sorbo
		var data = makeFindItem(aItemFilter, aCount, start, end);

		var req = new calExchangeRequest(this, data,
						 "findItemReply");

		req.mListener = aListener;
		req.mOp = Components.interfaces.calIOperationListener.GET;
		req.mInvite = wantInvitations;

		this.queueRequest(req);
	},

	completeGetRequest: function(aListener)
	{
		if (aListener == null)
			aListener = this.mReq.mListener;

		aListener.onOperationComplete(this,
		      Components.results.NS_OK,
		      Components.interfaces.calIOperationListener.GET,
		      null,
		      null);
	},

	queueRequest: function(aReq)
	{
		sdbg("doRequest\n");
		this.mReqQueue.push(aReq);
		this.kick();
	},

	kick: function()
	{
		if (this.mInit != 2)
			return;

		if (this.mRunning
		    || this.mReqQueue.length == 0)
			return;

		this.mRunning = true;

		var req = this.mReqQueue.shift();

		this.doRequest(req);
	},

	autodiscover: function()
	{
		// XXX should do:
		// 1) domain/autodiscover/autodiscover.xml
		// 2) autodiscover.domain/autodiscover/autodiscover.xml

		sdbg("autodiscover\n");

		var parts = this.mUser.split("@");
		var url = "http://autodiscover." 
			  + parts[1] 
			  + "/autodiscover/autodiscover.xml";

		this.setupReq(url);

		var data = makeAutodiscover(this.mUser);
		var req  = new calExchangeRequest(this, data,
						  "autodiscoverReply");

		this.mReqQueue.unshift(req);
		this.kick();
	},

	doRequest: function(aReq)
	{
		this.mReq = aReq;

		if (this.mServer != null)
			this.setupReq(this.mServer);

		sdbg("Sending: " + aReq.mData + "\n");
		this.mXmlReq.send(aReq.mData);
	},

	reply: function()
	{
		var req = this.mXmlReq;

		if (req.readyState != 4)
			return;

		if (req.status == 302) {
			var loc = this.mHttp.getResponseHeader("Location");

			sdbg("Redirect: " + loc + "\n");

			// XXX pheer loops.
			this.mXmlReq.abort();
			this.newXmlReq();
			this.setupReq(loc);
			this.doRequest(this.mReq);

			return;
		}

		sdbg("reply\n");

		sdbg(req.responseText + "\n");

		this.mAuthFail = 0;
		this.mRunning  = false;

		if (!this.isError(req))	
			this[this.mReq.mCb](req.responseXML);

		this.kick();
	},

	isError: function(aReq)
	{
		if (aReq.status != 200) {
			if (aReq.status == 403)
				this.discardAccount();
			// XXX parse it
			this.fail(aReq.responseText);
			return true;
		}

		var aReply = aReq.responseXML;

		var err = getXml(aReply, "ResponseCode", _msm);

		if (err == "" || err == "NoError")
			return false;

		var msg = getXml(aReply, "MessageText", _msm); 

		sdbg("Error: " + err + " msg: " + msg + "\n");

		this.fail(msg);
		return true;
	},

	newXmlReq: function()
	{
		this.mXmlReq = 
		  Components.classes ["@mozilla.org/xmlextras/xmlhttprequest;1"]
		   .createInstance(Components.interfaces.nsIJSXMLHttpRequest);
	},

	fail: function(aMsg)
	{
		sdbg("fail status: " + this.mXmlReq.status + " " + aMsg + "\n");

		var req  = this.mReq;

		// XXX force prompt
		const calIErrors = Components.interfaces.calIErrors;
		var code = calIErrors.MODIFICATION_FAILED;

		if (req.mAvail != null) {
			req.mListener.onResult(null, null);

			var xml = this.mXmlReq.responseXML;
			var err = getXml(xml, "ResponseCode", _msm);

			// XXX should avoid the second one
			if (err == "ErrorMailRecipientNotFound"
			    || err == "ErrorInvalidSmtpAddress")
				return true;
		} else {
			req.mListener.onOperationComplete(this,
                      					  code,
		      					  req.mOp,
		      					  null,
		      					  aMsg);
		}

		this.mObservers.notify("onError",
				       [this, code, aMsg]);

		return true;
	},

	setupReq: function(aUrl)
	{
		this.mXmlReq.open("POST", aUrl, true, this.mUser, this.mPass);

		var tmp = this;

		this.mXmlReq.onreadystatechange = function() {
			tmp.reply();
		}

		this.mXmlReq.overrideMimeType('text/xml');
		this.mXmlReq.setRequestHeader("Content-Type", "text/xml");

		this.mXmlReq.channel.notificationCallbacks = this;

		var httpChannel = this.mXmlReq.channel.QueryInterface(
			Components.interfaces.nsIHttpChannel);

		// XXX we want to preserve POST across 302 redirects
		httpChannel.redirectionLimit = 0;

		this.mHttp = httpChannel;
	},

	autodiscoverReply: function(aReply)
	{
		sdbg("autodiscoverreply");
		this.mServer = getXml(aReply, "ASUrl");

		ASSERT(this.mServer != "");

		sdbg("Autodiscover server: " + this.mServer + "\n");
	},

	findItemReply: function(aReply)
	{
		// FindItem does not return body.  Calendar's GetItem never
		// seems to be called, so we gotta be proactive and call
		// Exchange's getItem to retrieve body.

		var ids = aReply.getElementsByTagNameNS(_mst, "ItemId");

		sdbg("findItemReply: " + ids.length + "\n");

		if (ids.length == 0) {
			this.completeGetRequest();
			return;
		}

		var stuff = new Array();
		var i = 0;
		for (i = 0; i < ids.length; i++) {
			var item = {};
			item.id  = ids[i].getAttribute("Id");
			item.key = ids[i].getAttribute("ChangeKey");
			stuff.push(item);
		}

		var data = makeGetItem(stuff);
		var req  = new calExchangeRequest(this, data, "getItemReply");
		req.mListener = this.mReq.mListener;
		req.mOp = this.mReq.mOp;
		req.mInvite = this.mReq.mInvite;

		this.queueRequest(req);
	},

	getItemReply: function(aReply)
	{
		sdbg("getItemReply\n");

		var el = this.mReq.mInvite ? "MeetingRequest" : "CalendarItem";

		var items = aReply.getElementsByTagNameNS(_mst, el);

		var i = 0;
		for (i = 0; i < items.length; i++) {
			this.doGetItemReply(items[i]);
		}

		this.completeGetRequest();
	},

	doGetItemReply: function(aItem)
	{
		var item = createEvent();
		item.calendar = this;

		var id = aItem.getElementsByTagNameNS(_mst, "ItemId")[0];
		item.id = id.getAttribute("Id");

		item.setProperty("SEQUENCE", id.getAttribute("ChangeKey"));

		// XXX ugly
		item.removeAllAttendees();

		var i = 0;
		for (i = 0; i < _fields.length; i++) {
			var f = _fields[i];

			f.mCal = this; // XXX
			f.fillCal(item, aItem);
		}

		item.setProperty("X-SOR-INVITE", this.mReq.mInvite);

		var listener = this.mReq.mListener;

		listener.onGetResult(this,
				     Components.results.NS_OK,
				     Components.interfaces.calIEvent,
				     null,
				     1,
				     [item]);
	},

	modifyItem: function(aNewItem, aOldItem, aListener)
	{
		sdbg("modifyItem\n");

		var data;
		var del = false;

		var invite = aNewItem.getProperty("X-SOR-INVITE") || this.isInvitation(aNewItem);

		if (invite === true) {
			var me = this.getInvitedAttendee(aNewItem);

			const m = {
				// XXX
				"NEEDS-ACTION"	: "t:TentativelyAcceptItem",
				"TENTATIVE"	: "t:AcceptItem",

				"ACCEPTED"	: "t:AcceptItem",
				"DECLINED"	: "t:DeclineItem"
			};

			if (me.participationStatus == "DECLINED")
				del = true;

			sdbg("PS: " + me.participationStatus + "\n");

			data = makeMeetingResponse(aNewItem,
				m[me.participationStatus]);

			// XXX
			var a = this.getInvitedAttendee(aOldItem)
					.participationStatus;
			var b = me.participationStatus;
			if (((a == b) 
				&& (a == "ACCEPTED" || a == "NEEDS-ACTION"))
			    || ((a == "TENTATIVE") && (b == "NEEDS-ACTION")))
				data = "";
		} else
			data = makeUpdateItem(aNewItem, aOldItem);

		if (data == "") {
			this.completeUpdateRequest(aListener, aNewItem,
						   aOldItem);

			return;
		}

		var req = new calExchangeRequest(this, data,
						 "updateItemReply");

		req.mListener = aListener;
		req.mNewItem  = aNewItem;
		req.mOldItem  = aOldItem;
		req.mOp = Components.interfaces.calIOperationListener.MODIFY;
		req.mDel = del;

		this.queueRequest(req);
	},

	updateItemReply: function(aReply)
	{
		sdbg("updateItemReply\n");

		// XXX parse new item from XML

		var listener = this.mReq.mListener;
		var item = this.mReq.mNewItem.clone();

		var id = aReply.getElementsByTagNameNS(_mst, "ItemId")
		if (id.length != 0) {
			item.setProperty("SEQUENCE", 
					 id[0].getAttribute("ChangeKey"));

			item.id = id[0].getAttribute("Id");
		}

		this.completeUpdateRequest(listener, item, this.mReq.mOldItem);

		if (this.mReq.mDel)
			this.mObservers.notify("onDeleteItem", [item]);
	},

	completeUpdateRequest: function(aListener, aNewItem, aOldItem)
	{
		aListener.onOperationComplete(this,
                      Components.results.NS_OK,
                      Components.interfaces.calIOperationListener.MODIFY,
                      aNewItem.id,
                      aNewItem);

		this.mObservers.notify("onModifyItem",
				       [aNewItem, aOldItem]);
	},

	deleteItem: function(aItem, aListener)
	{
		sdbg("deleting item\n");

		var data = makeDeleteItem(aItem);

		var req = new calExchangeRequest(this, data,
						 "deleteItemReply");

		req.mListener = aListener;
		req.mItem     = aItem;
		req.mOp = Components.interfaces.calIOperationListener.DELETE;

		this.queueRequest(req);
	},

	deleteItemReply: function(aReply)
	{
		sdbg("deleteItemReply\n");

		var item     = this.mReq.mItem;
		var listener = this.mReq.mListener;

		listener.onOperationComplete(this,
                      Components.results.NS_OK,
                      Components.interfaces.calIOperationListener.DELETE,
                      item.id,
                      item);

		this.mObservers.notify("onDeleteItem", [item]);
	},

	addItem: function(aItem, aListener)
	{
		return this.adoptItem(aItem.clone(), aListener);
	},

	adoptItem: function(aItem, aListener)
	{
		sdbg("adding item\n");

		var data = makeCreateItem(aItem);

		var req = new calExchangeRequest(this, data,
						 "createItemReply");

		req.mListener = aListener;
		req.mItem     = aItem;
		req.mOp = Components.interfaces.calIOperationListener.ADD;

		this.queueRequest(req);
	},

	createItemReply: function(aReply)
	{
		sdbg("createItemReply\n");

		var item     = this.mReq.mItem;
		var listener = this.mReq.mListener;

		var id  = aReply.getElementsByTagNameNS(_mst, "ItemId")[0];
		item.id = id.getAttribute("Id");
		item.setProperty("SEQUENCE", id.getAttribute("ChangeKey"));

		listener.onOperationComplete(this,
                      Components.results.NS_OK,
                      Components.interfaces.calIOperationListener.ADD,
                      item.id,
                      item);

		this.mObservers.notify("onAddItem", [item]);
	},

	getFreeBusyIntervals: function(aCalId, aRangeStart, aRangeEnd,
				       aBusyTypes, aListener)
	{
		sdbg("getFreeBusyIntervals: " + aCalId + "\n");

		if (aCalId.indexOf("@") < 0 || aCalId.indexOf(".") < 0) {
            		// No valid email, screw it
			aListener.onResult(null, null);
			return null;
		}

		var tz = aRangeStart.timezone;

		var data = makeGetUserAvailability(
					aCalId.replace(/^MAILTO:/, ""),
					aRangeStart, aRangeEnd);

		var req = new calExchangeRequest(this, data,
						 "UserAvailabilityReply");

		req.mListener = aListener;
		req.mAvail    = true;
		req.mCalId    = aCalId;
		req.mTz       = tz;

		this.queueRequest(req);
	},

	UserAvailabilityReply: function(aReply)
	{
		sdbg("UserAvailabilityReply\n");

		var items = new Array();

		var ci = aReply.getElementsByTagNameNS(_mst, "CalendarEvent");
		for (i = 0; i < ci.length; i++) {
			var item = this.doAvailability(ci[i]);
			items.push(item);
		}

		this.mReq.mListener.onResult(null, items);
	},

	doAvailability: function(aCi)
	{
		const x = Components.interfaces.calIFreeBusyInterval;

		const types = {
			"Free" 		: x.FREE,
			"Tentative"	: x.BUSY_TENTATIVE,
			"Busy"		: x.BUSY,
			"OOF"		: x.BUSY_UNAVAILABLE,
			"NoData"	: x.UNKNOWN
		};

		var tz = this.mReq.mTz;

		var start = _f["fromRFC3339"](getXml(aCi, "StartTime", _mst),
					      tz);
		var end   = _f["fromRFC3339"](getXml(aCi, "EndTime", _mst), tz);
		var type  = types[getXml(aCi, "BusyType", _mst)];

		return new _f["FreeBusyInterval"](this.mReq.mCalId, type,
					       	  start, end);
	},

	// XXX seem to get auth prompts for NTLM.  Suppress them here.
	getInterface: function(iid)
	{
		sdbg("getInterface\n");

		if (iid.equals(Components.interfaces.nsIAuthPrompt)) {
			if (this.mAuthFail++ > 2) {
				this.mAuthFail = 0;
				this.discardAccount();
			}

			return fakePrompt(this);
		} 

		sdbg("IID: " + iid + "\n");

//		throw Components.results.NS_ERROR_NO_INTERFACE;
		return null;
	},

	getAccount: function()
	{
		var save = true;

		sdbg("getAccount\n");

                var username = { value: this.mUser };
                var password = { value: this.mPass };
                var persist  = { value: false };
		var title = "Microsoft Exchange";
		var realm = this.mUri.prePath;

		var got = _f["passwordManagerGet"]
				(this.mUser, password, realm, realm);

		if (got) {
			this.mPass = password.value;
			return;
		}

		_f["getCredentials"](title, realm, username, password, persist);

		this.mUser = username.value;
		this.mPass = password.value;

		//sdbg("PASS: " + this.mPass + "\n");

		if (persist.value) {
			_f["passwordManagerSave"](this.mUser, this.mPass,
						  realm, realm);
		}
	},

	discardAccount: function()
	{
		var realm = this.mUri.prePath;

		sdbg("discarding password info on " + realm + " / " + this.mUser + "\n");

		_f["passwordManagerRemove"](this.mUser, realm, realm);
		getAccount();
	},

	mReqQueue: null,
	mServer: null,
	mUri: null,
	mRunning: null,
	mUser: null,
	mPass: null,
	mReq: null,
	mXmlReq: null,
	mInit: 0,
	mAuthFail: 0
};

function calExchangeRequest(aCal, aData, aCb)
{
	this.wrappedJSObject = this;
	this.mCal  = aCal;
	this.mData = aData;
	this.mCb   = aCb;
}

calExchangeRequest.prototype = {
	QueryInterface: function(aIID) {
		return doQueryInterface(this,
					calExchangeRequest.prototype,
					aIID,
					null,
					g_classInfo["calExchangeRequest"]);
	},

	mCal: null,
	mData: null,
	mCb: null,
	mListener: null
};

function makeFindItem(aItemFilter, aCount, aRangeStart, aRangeEnd)
{
	var data = "";
	var folder = "calendar";
	var max = "";

	var wantInvitations = ((aItemFilter & Components.interfaces.calICalendar
		  		.ITEM_FILTER_REQUEST_NEEDS_ACTION) != 0);

	if (aCount != 0)
        	max = ' MaxEntriesReturned="' + aCount + '"';

	var view = '<CalendarView'
               	   + ' StartDate="' + aRangeStart + '"'
                   + '  EndDate="' + aRangeEnd + '"'
                   + max
                   + '/>';

	if (aRangeEnd == "")
		view = "";

	// XXX search for date
	if (wantInvitations) {
		folder = "inbox";

		view = "";
		view += '<Restriction>';
		view += '  <t:IsEqualTo>';
		view += '    <t:FieldURI FieldURI="item:ItemClass"/>';
		view += '      <t:FieldURIOrConstant>';
		view += '        <t:Constant Value="IPM.Schedule.Meeting.Request"/>';
		view += '      </t:FieldURIOrConstant>';
		view += '    </t:IsEqualTo>';
		view += '</Restriction>';
	}

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <FindItem xmlns="http://schemas.microsoft.com/exchange/services/2006/messages"';
	data += '      xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types"';
	data += '      Traversal="Shallow">';
	data += '      <ItemShape>';
	data += '        <t:BaseShape>IdOnly</t:BaseShape>';
	data += '      </ItemShape>';
	data += view;
	data += '      <ParentFolderIds>';
	data += '        <t:DistinguishedFolderId Id="' + folder + '"/>';
	data += '      </ParentFolderIds>';
	data += '    </FindItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeGetItem(aItems)
{
	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <GetItem xmlns="http://schemas.microsoft.com/exchange/services/2006/messages"';
	data += '      xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '      <ItemShape>';
	data += '        <t:BaseShape>IdOnly</t:BaseShape>';
	data += '        <t:BodyType>Text</t:BodyType>';
	data += '        <t:AdditionalProperties>';

	for (i in _fields) {
		var f = _fields[i];

		data += '<t:FieldURI FieldURI="' + f.mExType + '"/>';
	}

	data += '        </t:AdditionalProperties>';
	data += '      </ItemShape>';
	data += '      <ItemIds>';

	for (i in aItems) {
		data += '<t:ItemId Id="' + aItems[i].id + '"'
			+ ' ChangeKey="' + aItems[i].key + '"/>';
	}

	data += '      </ItemIds>';
	data += '    </GetItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeAutodiscover(aEmail)
{
	var data = "";

        data += '<?xml version="1.0" encoding="utf-8"?>';
        data += '<Autodiscover xmlns="http://schemas.microsoft.com/exchange/autodiscover/outlook/requestschema/2006">';
        data += '  <Request>';
        data += '    <EMailAddress>' + aEmail + '</EMailAddress>';
        data += '    <AcceptableResponseSchema>http://schemas.microsoft.com/exchange/autodiscover/outlook/responseschema/2006a</AcceptableResponseSchema>';
        data += '  </Request>';
        data += '</Autodiscover>';

	return data;
}

function sendMail(aItem)
{
	var inv = aItem.getProperty("X-MOZ-SEND-INVITATIONS");

	if (inv == "TRUE" || true)
		return "SendToAllAndSaveCopy";

	return "SendToNone";
}

function makeUpdateItem(aNewItem, aOldItem)
{
	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <UpdateItem MessageDisposition="SaveOnly"';
	data += '      ConflictResolution="AutoResolve"';
	data += '      SendMeetingInvitationsOrCancellations="' + sendMail(aNewItem) + '"';
	data += '      xmlns="http://schemas.microsoft.com/exchange/services/2006/messages"';
	data += '      xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '      <ItemChanges>';
	data += '        <t:ItemChange>';
	data += '          <t:ItemId Id="' + aOldItem.id + '"'
			   + ' ChangeKey="' + aOldItem.getProperty("SEQUENCE") + '"/>';
        data += '          <t:Updates>';

	var up = "";

	for (i in _fields) {
		var f = _fields[i];

		up += f.doUpdate(aNewItem, aOldItem);
	}

	if (up == "")
		return "";

	data += up;

	data += '	   </t:Updates>';
	data += '	 </t:ItemChange>';
	data += '      </ItemChanges>';
	data += '    </UpdateItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeDeleteItemField(aField)
{
	return makeXItemField(aField, "t:DeleteItemField", "");
}

function makeSetItemField(aField, aContent)
{
	aContent = "<t:CalendarItem>" + aContent + "</t:CalendarItem>";

	return makeXItemField(aField, "t:SetItemField", aContent);
}

function makeXItemField(aField, aOperation, aContent)
{
	var data = "";

	data += '<' + aOperation + '>';
	data += '  <t:FieldURI FieldURI="' + aField + '"/>';
	data += aContent;
	data += '</' + aOperation + '>';

	return data;
}

function makeCreateItem(aItem)
{
	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <CreateItem SendMeetingInvitations="' + sendMail(aItem) + '"';
	data += '                xmlns="http://schemas.microsoft.com/exchange/services/2006/messages"';
	data += '                xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '      <SavedItemFolderId>';
	data += '        <t:DistinguishedFolderId Id="calendar"/>';
	data += '      </SavedItemFolderId>';
	data += '      <Items>';
	data += '        <t:CalendarItem xmlns="http://schemas.microsoft.com/exchange/services/2006/types">';

	for (i in _fields) {
		var f = _fields[i];

		data += f.doCreate(aItem);
	}

	data += '        </t:CalendarItem>';
	data += '      </Items>';
	data += '    </CreateItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeMeetingResponse(aItem, aResp)
{
	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <CreateItem SendMeetingInvitations="' + sendMail(aItem) + '"';
	data += '    		 MessageDisposition="SendAndSaveCopy"';
	data += '                xmlns="http://schemas.microsoft.com/exchange/services/2006/messages"';
	data += '                xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '      <Items>';
	data += '	<' + aResp + '>';
	data += '	 <t:ReferenceItemId Id="' + aItem.id 
			+ '" ChangeKey="' + aItem.getProperty("SEQUENCE") + '" />';
	data += '	</' + aResp + '>';
	data += '      </Items>';
	data += '    </CreateItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeTag(aTag, aContent, aAttrib)
{
	var data = "";

	if (aAttrib == null)
		aAttrib = "";

	data += '<' + aTag + ' ' + aAttrib + '>';
	data += "<![CDATA[" + aContent + "]]>";
	data += '</' + aTag + '>';

	return data;
}

function makeTag2(aTag, aContent)
{
	return "<" + aTag + ">" + aContent + "</" + aTag + ">";
}

function makeDeleteItem(aItem)
{
	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <DeleteItem DeleteType="HardDelete"';
	data += '      SendMeetingCancellations="' + sendMail(aItem) + '"';
	data += '                xmlns="http://schemas.microsoft.com/exchange/services/2006/messages">';
	data += '      <ItemIds>';
	data += '        <t:ItemId Id="' + aItem.id;
        data += '"         ChangeKey="' + aItem.getProperty("SEQUENCE") + '"/>';
	data += '      </ItemIds>';
	data += '    </DeleteItem>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function makeGetUserAvailability(aId, aStart, aEnd)
{
	// XXX this is fucked up - need to code it properly
	// microsoft got it wrong - server should have TZ info, client should
	// supply ID, or at least have that option. =(
	var tz = aStart.timezone;

	var standard = tz.icalComponent.getFirstSubcomponent("STANDARD");

	aStart = getDate(aStart);
	aEnd   = getDate(aEnd);

	var b = off2min(getIcalString(standard, "TZOFFSETTO"));

	sdbg("OFF: " + b + "\n");

	var data = "";

	data += '<?xml version="1.0" encoding="utf-8"?>';
	data += '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"';
	data += '  xmlns:t="http://schemas.microsoft.com/exchange/services/2006/types">';
	data += '  <soap:Body>';
	data += '    <GetUserAvailabilityRequest xmlns="http://schemas.microsoft.com/exchange/services/2006/messages">';

	data += '      <t:TimeZone>';
	data += makeTag2("t:Bias", b);
	data += doTz("t:StandardTime", standard, 0);
	data += doTz("t:DaylightTime",
		     tz.icalComponent.getFirstSubcomponent("DAYLIGHT"));
	data += '      </t:TimeZone>';

	data += '      <MailboxDataArray>';
	data += '        <t:MailboxData>';
	data += '          <t:Email>';
	data += '            <t:Address>' + aId + '</t:Address>';
	data += '          </t:Email>';
	data += makeTag2("t:AttendeeType", "Required");
	data += '        </t:MailboxData>';
	data += '      </MailboxDataArray>';

	data += '      <t:FreeBusyViewOptions>';

	if (aStart != "" || aEnd != "") {
		data += '<t:TimeWindow>';

		if (aStart != "")
			data += '<t:StartTime>' + aStart + '</t:StartTime>';

		if (aEnd != "")
			data += '<t:EndTime>' + aEnd + '</t:EndTime>';

		data += '</t:TimeWindow>';
	}

	data += makeTag2("t:MergedFreeBusyIntervalInMinutes", 60);
	data += '      </t:FreeBusyViewOptions>';

	data += '    </GetUserAvailabilityRequest>';
	data += '  </soap:Body>';
	data += '</soap:Envelope>';

	return data;
}

function getIcalString(component, property)
{
	var prop = (component && component.getFirstProperty(property));
	return (prop ? prop.valueAsIcalString : null);
}

function off2min(aOff)
{
	var min = parseInt(aOff.substring(3, 5), 10);

	min += parseInt(aOff.substring(1, 3), 10) * 60;

	// bah... tell m$
	if (aOff[0] == '+')
		min *= -1;

	return min;
}

// XXX FIXME
function doTz(aTag, aTz, aBias)
{
	var xml     = "";
	var time    = "02:00:00";
	var dayoff  = "1";
	var month   = "1";
	var dayname = "Monday";
	var bias    = "0";

	sdbg("ATZ: " + aTz  + "\n");	

	function getStuff(rule, tok) {
		var crap = rule.split(";");

		for (i in crap) {
			var parts = crap[i].split("=");

			if (parts[0] == tok)
				return parts[1];
		}

		return null;
	}

	const daymap = {
		"MO" : "Monday",
		"TU" : "Tuesday",
		"WE" : "Wednesday",
		"TH" : "Thursday",
		"FR" : "Friday",
		"SA" : "Saturday",
		"SU" : "Sunday"
	};

	function doIt() {
		bias = aBias;

		if (bias == null) {
			var a = off2min(getIcalString(aTz, "TZOFFSETFROM"));
			var b = off2min(getIcalString(aTz, "TZOFFSETTO"));

			bias = b - a;
		}

		var rule = getIcalString(aTz, "RRULE");
		if (rule == null)
			return;

		var day  = getStuff(rule, "BYDAY");

		dayoff  = parseInt(day.substring(0, day.length - 2), 10);
		dayname = daymap[day.substring(day.length - 2, day.length)];

		// XXX
		if (dayoff < 0) {
			dayoff = 6 + dayoff;
		}

		month = getStuff(rule, "BYMONTH")
	}

	// m$ really got it wrong.  We need to spoof timezone info!
	if (aTz != null) {
		month = "10";
		doIt();
	}

	xml += "<" + aTag + ">";

	xml += makeTag2("t:Bias", bias);
	xml += makeTag2("t:Time", time);
	xml += makeTag2("t:DayOrder", dayoff);
	xml += makeTag2("t:Month", month);
	xml += makeTag2("t:DayOfWeek", dayname);

	xml += "</" + aTag + ">";

	return xml;
}

function getXml(aXml, aName, aNs)
{
	var items;

	if (aNs == null)
		items = aXml.getElementsByTagName(aName);
	else
		items = aXml.getElementsByTagNameNS(aNs, aName);
	
	if (items.length == 0)
		return "";

	return items[0].textContent;
}

function doXPath(aXml, aXPath)
{
	var doc = aXml.ownerDocument;

	var res = function(aPref) {
		var ns = {
			't' : _mst
		};

		return ns[aPref] || null;
	}

	return doc.evaluate(aXPath, aXml, res,
		     Components.interfaces.nsIDOMXPathResult.ANY_TYPE, null);
}

function fakePrompt(aCal)
{
	sdbg("making prompt\n");

	var p = {
		promptUsernameAndPassword: function(dialogTitle,
						    text,
						    passwordRealm,
						    savePassword,
						    user,
						    pwd)
		{
			user.value = aCal.mUser;
			pwd.value  = aCal.mPass;

			return true;
		}
	}

	return p;
}

function getDate(aRange)
{
	// Requesting only a DATE returns items based on UTC. Therefore,
	// we make sure both start and end dates include a time and
	// timezone. This may not quite be what was requested, but I'd
	// say its a shortcoming of rfc3339.  
	//
	// [ripped comment from gdata - like many other things ;D]
	if (aRange) {
		aRange = aRange.clone();
		aRange.isDate = false;
	}

	return _f["toRFC3339"](aRange);
}

function sdbg(msg)
{
	return;
	dump(msg);
}

init_fields();
